@@ -1,5 +1,5 @@
 /* libavl - manipulates AVL trees.
-   Copyright (C) 1998 Free Software Foundation, Inc.
+   Copyright (C) 1998, 1999 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -60,9 +60,11 @@ static void *
 xmalloc (size_t size)
 {
   void *vp;
+
   if (size == 0)
     return NULL;
   vp = malloc (size);
+
   assert (vp != NULL);
   if (vp == NULL)
     {
@@ -82,7 +84,7 @@ avl_create (MAYBE_ARENA avl_comparison_func cmp, void *param)
 {
   avl_tree *tree;
 
-  assert (cmp);
+  assert (cmp != NULL);
 #if PSPP
   if (owner)
     tree = arena_alloc (owner, sizeof (avl_tree));
@@ -113,7 +115,7 @@ avl_create (MAYBE_ARENA avl_comparison_func cmp, void *param)
 void
 avl_destroy (avl_tree *tree, avl_node_func free_func)
 {
-  assert (tree);
+  assert (tree != NULL);
   
 #if PSPP
   if (free_func || tree->owner == NULL)
@@ -181,7 +183,7 @@ avl_free (avl_tree *tree)
 int
 avl_count (const avl_tree *tree)
 {
-  assert (tree);
+  assert (tree != NULL);
   return tree->count;
 }
 
@@ -231,7 +233,7 @@ avl_copy (MAYBE_ARENA const avl_tree *tree, avl_copy_func copy)
   avl_node **qp = qa;		/* Stack QA: stack pointer. */
   avl_node *q;
   
-  assert (tree);
+  assert (tree != NULL);
 #if PSPP
   new_tree = avl_create (owner, tree->cmp, tree->param);
 #else
@@ -394,9 +396,11 @@ avl_probe (avl_tree *tree, void *item)
      random insertions.  */
 
   /* A1. */
-  avl_node *t = &tree->root;
+  avl_node *t;
   avl_node *s, *p, *q, *r;
-  assert (tree);
+  
+  assert (tree != NULL);
+  t = &tree->root;
   s = p = t->link[0];
 
   if (s == NULL)
@@ -566,12 +570,11 @@ avl_find (const avl_tree *tree, const void *item)
 {
   const avl_node *p;
 
-  assert (tree);
+  assert (tree != NULL);
   for (p = tree->root.link[0]; p; )
     {
       int diff = tree->cmp (item, p->data, tree->param);
 
-      /* A3. */
       if (diff < 0)
 	p = p->link[0];
       else if (diff > 0)
@@ -583,6 +586,37 @@ avl_find (const avl_tree *tree, const void *item)
   return NULL;
 }
 
+/* Search TREE for an item close to the value of ITEM, and return it.
+   This function will return a null pointer only if TREE is empty. */
+void *
+avl_find_close (const avl_tree *tree, const void *item)
+{
+  const avl_node *p;
+
+  assert (tree != NULL);
+  p = tree->root.link[0];
+  if (p == NULL)
+    return NULL;
+  
+  for (;;)
+    {
+      int diff = tree->cmp (item, p->data, tree->param);
+      int t;
+
+      if (diff < 0)
+	t = 0;
+      else if (diff > 0)
+	t = 1;
+      else
+	return p->data;
+
+      if (p->link[t])
+	p = p->link[t];
+      else
+	return p->data;
+    }
+}
+
 /* Searches AVL tree TREE for an item matching ITEM.  If found, the
    item is removed from the tree and the actual item found is returned
    to the caller.  If no item matching ITEM exists in the tree,
@@ -604,7 +638,7 @@ avl_delete (avl_tree *tree, const void *item)
   avl_node **q;
   avl_node *p;
 
-  assert (tree);
+  assert (tree != NULL);
 
   a[0] = 0;
   pa[0] = &tree->root;
@@ -818,7 +852,7 @@ avl_insert (avl_tree *tree, void *item)
 {
   void **p;
   
-  assert (tree);
+  assert (tree != NULL);
   
   p = avl_probe (tree, item);
   return (*p == item) ? NULL : *p;
@@ -833,7 +867,7 @@ avl_replace (avl_tree *tree, void *item)
 {
   void **p;
 
-  assert (tree);
+  assert (tree != NULL);
   
   p = avl_probe (tree, item);
   if (*p == item)
@@ -852,7 +886,7 @@ void *
 (avl_force_delete) (avl_tree *tree, void *item)
 {
   void *found = avl_delete (tree, item);
-  assert (found);
+  assert (found != NULL);
   return found;
 }
 
@@ -1027,23 +1061,29 @@ compare_trees (avl_node *a, avl_node *b)
    places in order to be able to see what's really going on.  Also,
    memory debuggers like Checker or Purify are very handy. */
 #define TREE_SIZE 1024
+#define N_ITERATIONS 16
 int
 main (int argc, char **argv)
 {
   int array[TREE_SIZE];
   int seed;
+  int iteration;
   
   if (argc == 2)
     seed = atoi (argv[1]);
   else
     seed = time (0) * 257 % 32768;
 
-  for (;;)
+  fputs ("Testing avl...\n", stdout);
+  
+  for (iteration = 1; iteration <= N_ITERATIONS; iteration++)
     {
       avl_tree *tree;
       int i;
       
-      printf ("Random number seed: %d\n", seed);
+      printf ("Iteration %4d/%4d: seed=%5d", iteration, N_ITERATIONS, seed);
+      fflush (stdout);
+      
       srand (seed++);
 
       for (i = 0; i < TREE_SIZE; i++)
@@ -1070,10 +1110,11 @@ main (int argc, char **argv)
 
 	  if (i % 128 == 0)
 	    {
-	      fputc ('.', stdout);
+	      putchar ('.');
 	      fflush (stdout);
 	    }
 	}
+      fputs (" good.\n", stdout);
 
       avl_destroy (tree, NULL);
     }
