@@ -1,5 +1,5 @@
 /* libavl - manipulates AVL trees.
-   Copyright (C) 1998 Free Software Foundation, Inc.
+   Copyright (C) 1998, 1999 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -80,7 +80,7 @@ avlt_create (avl_comparison_func cmp, void *param)
 {
   avlt_tree *tree;
 
-  assert (cmp);
+  assert (cmp != NULL);
   tree = xmalloc (sizeof (avlt_tree));
 
   tree->root.link[0] = tree->root.link[1] = &tree->root;
@@ -104,7 +104,7 @@ avlt_create (avl_comparison_func cmp, void *param)
 void
 avlt_destroy (avlt_tree *tree, avl_node_func free_func)
 {
-  assert (tree);
+  assert (tree != NULL);
   
   if (tree->root.link[0] != &tree->root)
     {
@@ -168,7 +168,7 @@ avlt_free (avlt_tree *tree)
 int
 avlt_count (const avlt_tree *tree)
 {
-  assert (tree);
+  assert (tree != NULL);
   return tree->count;
 }
 
@@ -190,7 +190,7 @@ avlt_copy (const avlt_tree *tree, avl_copy_func copy)
   const avlt_node *p;
   avlt_node *q;
   
-  assert (tree);
+  assert (tree != NULL);
   new_tree = avlt_create (tree->cmp, tree->param);
   new_tree->count = tree->count;
   p = &tree->root;
@@ -281,7 +281,7 @@ avlt_thread (struct avl_tree *_tree)
   /* Algorithm L's variables. */
   avlt_node *p, *q;
 
-  assert (tree);
+  assert (tree != NULL);
 
   /* T1. */
   ap = an;
@@ -355,7 +355,7 @@ avlt_unthread (avlt_tree *tree)
   int ap = 0;				/* Stack A: height. */
   avlt_node *p;
 
-  assert (tree);
+  assert (tree != NULL);
   p = tree->root.link[0];
   if (p != &tree->root)
     for (;;)
@@ -406,11 +406,12 @@ avlt_unthread (avlt_tree *tree)
 void
 avlt_walk (const avlt_tree *tree, avl_node_func walk_func, void *param)
 {
-  const avlt_node *p = &tree->root;
+  const avlt_node *p;
 
   /* Uses Knuth's algorithm 2.3.1D (threaded inorder successor). */
   assert (tree && walk_func);
-  
+
+  p = &tree->root;
   for (;;)
     {
       if (p->tag[1] == MINUS)
@@ -478,7 +479,7 @@ avlt_next (const avlt_tree *tree, void **item)
 {
   const avlt_node *p;
 
-  assert (tree);
+  assert (tree != NULL);
   if (item == NULL)
     p = &tree->root;
   else
@@ -508,7 +509,7 @@ avlt_prev (const avlt_tree *tree, void **item)
 {
   const avlt_node *p;
 
-  assert (tree);
+  assert (tree != NULL);
   if (item == NULL)
     {
       /* Find node with greatest value. */
@@ -556,9 +557,11 @@ avlt_probe (avlt_tree *tree, void *item)
      the comparisons seen under random insertions.  */
 
   /* A1. */
-  avlt_node *t = &tree->root;
+  avlt_node *t;
   avlt_node *s, *p, *q, *r;
-  assert (tree);
+
+  assert (tree != NULL);
+  t = &tree->root;
   s = p = t->link[0];
 
   if (t->tag[0] == MINUS)
@@ -777,9 +780,14 @@ avlt_probe (avlt_tree *tree, void *item)
 void **
 avlt_find (const avlt_tree *tree, const void *item)
 {
-  const avlt_node *p = tree->root.link[0];
+  const avlt_node *p;
 
-  assert (tree);
+  assert (tree != NULL);
+  if (tree->root.tag[0] == MINUS)
+    /* Tree is empty. */
+    return NULL;
+
+  p = tree->root.link[0];
   for (;;)
     {
       int diff = tree->cmp (item, p->data, tree->param);
@@ -800,6 +808,39 @@ avlt_find (const avlt_tree *tree, const void *item)
     }
 }
 
+/* Search TREE for an item close to the value of ITEM, and return it.
+   This function will return a null pointer only if TREE is empty. */
+void **
+avlt_find_close (const avlt_tree *tree, const void *item)
+{
+  const avlt_node *p;
+
+  assert (tree != NULL);
+  if (tree->root.tag[0] == MINUS)
+    /* Tree is empty. */
+    return NULL;
+
+  p = tree->root.link[0];
+  for (;;)
+    {
+      int diff = tree->cmp (item, p->data, tree->param);
+      int t;
+
+      /* A3. */
+      if (diff < 0)
+	t = 0;
+      else if (diff > 0)
+	t = 1;
+      else
+	return (void **) &p->data;
+
+      if (p->tag[t] == PLUS)
+	p = p->link[t];
+      else
+	return (void **) &p->data;
+    }
+}
+
 /* Searches AVL tree TREE for an item matching ITEM.  If found, the
    item is removed from the tree and the actual item found is returned
    to the caller.  If no item matching ITEM exists in the tree,
@@ -820,7 +861,11 @@ avlt_delete (avlt_tree *tree, const void *item)
   
   avlt_node *p;
 
-  assert (tree);
+  assert (tree != NULL);
+
+  if (tree->root.tag[0] == MINUS)
+    /* Empty tree. */
+    return NULL;
 
   a[0] = 0;
   pa[0] = &tree->root;
@@ -1092,7 +1137,7 @@ avlt_insert (avlt_tree *tree, void *item)
 {
   void **p;
   
-  assert (tree);
+  assert (tree != NULL);
   
   p = avlt_probe (tree, item);
   return (*p == item) ? NULL : *p;
@@ -1107,7 +1152,7 @@ avlt_replace (avlt_tree *tree, void *item)
 {
   void **p;
 
-  assert (tree);
+  assert (tree != NULL);
   
   p = avlt_probe (tree, item);
   if (*p == item)
@@ -1126,13 +1171,13 @@ void *
 (avlt_force_delete) (avlt_tree *tree, void *item)
 {
   void *found = avlt_delete (tree, item);
-  assert (found);
+  assert (found != NULL);
   return found;
 }
 
 #if SELF_TEST
 
-/* Size of tree for stress testing. */
+/* Size of the tree used for testing. */
 #define TREE_SIZE 1024	
 
 /* Used to flag delayed aborting. */
@@ -1159,7 +1204,7 @@ print_structure (avlt_tree *tree, avlt_node *node, int level)
   char lc[] = "([{<`";
   char rc[] = ")]}>'";
 
-  assert (node);
+  assert (node != NULL);
   if (level >= 10)
     {
       printf ("Too deep, giving up.\n");
@@ -1377,7 +1422,7 @@ compare_trees (avlt_node *a, avlt_node *b)
 {
   int diff = 0;
   
-  assert (a != NULL && b != NULL);
+  assert (a && b);
   
   /* Separating these conditions makes it easier to pinpoint bad data
      under a memory debugger like Checker because each test is a
@@ -1431,23 +1476,29 @@ compare_trees (avlt_node *a, avlt_node *b)
    might want to insert some calls to print_structure() at strategic
    places in order to be able to see what's really going on.  Also,
    memory debuggers like Checker or Purify are very handy. */
+#define N_ITERATIONS 16
 int
 main (int argc, char **argv)
 {
   int array[TREE_SIZE];
   int seed;
+  int iteration;
   
   if (argc == 2)
     seed = atoi (argv[1]);
   else
     seed = time (0) * 257 % 32768;
 
-  for (;;)
+  fputs ("Testing avlt...\n", stdout);
+  
+  for (iteration = 1; iteration <= N_ITERATIONS; iteration++)
     {
       avlt_tree *tree;
       int i;
       
-      printf ("Random number seed: %d\n", seed);
+      printf ("Iteration %4d/%4d: seed=%5d", iteration, N_ITERATIONS, seed);
+      fflush (stdout);
+      
       srand (seed++);
 
       for (i = 0; i < TREE_SIZE; i++)
@@ -1477,10 +1528,11 @@ main (int argc, char **argv)
 
 	  if (i % 128 == 0)
 	    {
-	      fputc ('.', stdout);
+	      putchar ('.');
 	      fflush (stdout);
 	    }
 	}
+      fputs (" good.\n", stdout);
 
       avlt_destroy (tree, NULL);
     }
