@@ -1,5 +1,5 @@
 /* libavl - manipulates AVL trees.
-   Copyright (C) 1998 Free Software Foundation, Inc.
+   Copyright (C) 1998, 1999 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -83,7 +83,7 @@ avltr_create (avl_comparison_func cmp, void *param)
 {
   avltr_tree *tree;
 
-  assert (cmp);
+  assert (cmp != NULL);
   tree = xmalloc (sizeof (avltr_tree));
 
   tree->root.link[0] = NULL;
@@ -107,7 +107,7 @@ avltr_create (avl_comparison_func cmp, void *param)
 void
 avltr_destroy (avltr_tree *tree, avl_node_func free_func)
 {
-  assert (tree);
+  assert (tree != NULL);
   
   if (tree->root.link[0] != &tree->root)
     {
@@ -170,7 +170,7 @@ avltr_free (avltr_tree *tree)
 int
 avltr_count (const avltr_tree *tree)
 {
-  assert (tree);
+  assert (tree != NULL);
   return tree->count;
 }
 
@@ -192,7 +192,7 @@ avltr_copy (const avltr_tree *tree, avl_copy_func copy)
   const avltr_node *p;
   avltr_node *q;
   
-  assert (tree);
+  assert (tree != NULL);
   new_tree = avltr_create (tree->cmp, tree->param);
   new_tree->count = tree->count;
   p = &tree->root;
@@ -281,7 +281,7 @@ avltr_thread (struct avl_tree *_tree)
   /* Algorithm L's variables. */
   avltr_node *p, *q;
 
-  assert (tree);
+  assert (tree != NULL);
 
   /* T1. */
   ap = an;
@@ -347,7 +347,7 @@ avltr_unthread (avltr_tree *tree)
   int ap = 0;				/* Stack A: height. */
   avltr_node *p;
 
-  assert (tree);
+  assert (tree != NULL);
   p = tree->root.link[0];
   if (p != NULL)
     for (;;)
@@ -468,7 +468,7 @@ avltr_next (const avltr_tree *tree, void **item)
 {
   const avltr_node *p;
 
-  assert (tree);
+  assert (tree != NULL);
   if (item == NULL)
     p = &tree->root;
   else
@@ -505,9 +505,11 @@ avltr_probe (avltr_tree *tree, void *item)
      25% of the comparisons seen under random insertions.  */
 
   /* A1. */
-  avltr_node *t = &tree->root;
+  avltr_node *t;
   avltr_node *s, *p, *q, *r;
-  assert (tree);
+
+  assert (tree != NULL);
+  t = &tree->root;
   s = p = t->link[0];
 
   if (s == NULL)
@@ -562,7 +564,7 @@ avltr_probe (avltr_tree *tree, void *item)
 	      q->link[0] = NULL;
 	      break;
 	    }
-	  assert (q);
+	  assert (q != NULL);
 	}
       else
 	/* A2. */
@@ -725,9 +727,12 @@ avltr_probe (avltr_tree *tree, void *item)
 void **
 avltr_find (const avltr_tree *tree, const void *item)
 {
-  const avltr_node *p = tree->root.link[0];
+  const avltr_node *p;
 
-  assert (tree);
+  assert (tree != NULL);
+  p = tree->root.link[0];
+  if (p == NULL)
+    return NULL;
   for (;;)
     {
       int diff = tree->cmp (item, p->data, tree->param);
@@ -750,6 +755,40 @@ avltr_find (const avltr_tree *tree, const void *item)
     }
 }
 
+/* Search TREE for an item close to the value of ITEM, and return it.
+   This function will return a null pointer only if TREE is empty. */
+void **
+avltr_find_close (const avltr_tree *tree, const void *item)
+{
+  const avltr_node *p;
+
+  assert (tree != NULL);
+  p = tree->root.link[0];
+  if (p == NULL)
+    return NULL;
+  for (;;)
+    {
+      int diff = tree->cmp (item, p->data, tree->param);
+
+      /* A3. */
+      if (diff < 0)
+	{
+	  if (p->link[0])
+	    p = p->link[0];
+	  else
+	    return (void **) &p->data;
+	}
+      else if (diff > 0)
+	{
+	  if (p->rtag == MINUS)
+	    return (void **) &p->data;
+	  p = p->link[1];
+	}
+      else
+	return (void **) &p->data;
+    }
+}
+
 /* Searches AVL tree TREE for an item matching ITEM.  If found, the
    item is removed from the tree and the actual item found is returned
    to the caller.  If no item matching ITEM exists in the tree,
@@ -770,7 +809,7 @@ avltr_delete (avltr_tree *tree, const void *item)
   
   avltr_node *p;
 
-  assert (tree);
+  assert (tree != NULL);
 
   a[0] = 0;
   pa[0] = &tree->root;
@@ -1058,7 +1097,7 @@ avltr_insert (avltr_tree *tree, void *item)
 {
   void **p;
   
-  assert (tree);
+  assert (tree != NULL);
   
   p = avltr_probe (tree, item);
   return (*p == item) ? NULL : *p;
@@ -1073,7 +1112,7 @@ avltr_replace (avltr_tree *tree, void *item)
 {
   void **p;
 
-  assert (tree);
+  assert (tree != NULL);
   
   p = avltr_probe (tree, item);
   if (*p == item)
@@ -1092,14 +1131,14 @@ void *
 (avltr_force_delete) (avltr_tree *tree, void *item)
 {
   void *found = avltr_delete (tree, item);
-  assert (found);
+  assert (found != NULL);
   return found;
 }
 
 #if SELF_TEST
 
-/* Size of tree for stress testing. */
-#define TREE_SIZE 32
+/* Size of the tree used for testing. */
+#define TREE_SIZE 1024
 
 /* Used to flag delayed aborting. */
 int done = 0;
@@ -1324,7 +1363,7 @@ compare_trees (avltr_node *a, avltr_node *b)
 {
   int diff = 0;
   
-  assert (a != NULL && b != NULL);
+  assert (a && b);
   
   /* Separating these conditions makes it easier to pinpoint bad data
      under a memory debugger like Checker because each test is a
@@ -1378,23 +1417,29 @@ compare_trees (avltr_node *a, avltr_node *b)
    might want to insert some calls to print_structure() at strategic
    places in order to be able to see what's really going on.  Also,
    memory debuggers like Checker or Purify are very handy. */
+#define N_ITERATIONS 16
 int
 main (int argc, char **argv)
 {
   int array[TREE_SIZE];
   int seed;
+  int iteration;
   
   if (argc == 2)
     seed = atoi (argv[1]);
   else
     seed = time (0) * 257 % 32768;
 
-  for (;;)
+  fputs ("Testing avltr...\n", stdout);
+
+  for (iteration = 1; iteration <= N_ITERATIONS; iteration++)
     {
       avltr_tree *tree;
       int i;
       
-      printf ("Random number seed: %d\n", seed);
+      printf ("Iteration %4d/%4d: seed=%5d", iteration, N_ITERATIONS, seed);
+      fflush (stdout);
+      
       srand (seed++);
 
       for (i = 0; i < TREE_SIZE; i++)
@@ -1424,10 +1469,11 @@ main (int argc, char **argv)
 
 	  if (i % 128 == 0)
 	    {
-	      fputc ('.', stdout);
+	      putchar ('.');
 	      fflush (stdout);
 	    }
 	}
+      fputs (" good.\n", stdout);
 
       avltr_destroy (tree, NULL);
     }
